import asyncio


def create_queue_suppliers(settings=None):
    futures = {}

    async def queue_out(data, scope):
        sender = data.id
        entry = data.payload
        portal = data.portal

        recipient = entry["recipient"]
        action = entry["action"]

        persistence = scope.demand({"type": "persistence"})
        crypto = scope.demand({"type": "crypto"})

        recipient_state_entry = await persistence.get(sender + "<>" + recipient) or {}
        state = recipient_state_entry.get("value", {"queue": []})
        queue = state.get("queue", [])
        # print("Current state:", state)

        entry_index = state.get("out_index", 0) + 1
        entry_uuid = crypto["random_uuid"]()
        entry_id = entry_uuid
        # entry_id = json.dumps(
        #     [
        #         sender,
        #         recipient,
        #         action,
        #         entry_index,
        #         entry_uuid,
        #     ]
        # )

        future = asyncio.Future()
        futures[entry_id] = future

        queue_entry = {
            "id": entry_id,
            "index": entry_index,
            "sender": sender,
            "recipient": recipient,
            "action": action,
            "payload": entry["payload"],
        }

        queue.append(queue_entry)
        state["out_index"] = entry_index
        # print("New state:", state)
        await persistence.set(sender + "<>" + recipient, dict(state))

        asyncio.create_task(portal("queue.dispatch", {"recipient": recipient}))

        return await future

    async def queue_in(data, scope):
        # print("queue_in", data.payload)
        payload = data.payload["payload"]
        persistence = scope.demand({"type": "persistence"})

        sender = data.payload["sender"]
        sender_state_entry = await persistence.get(data.id + "<>" + sender) or {}
        state = sender_state_entry.get("value", {"queue": []})
        in_index = state.get("in_index", 0)
        init_id = state.get("init_id", None)
        out_queue = state.get("queue", [])

        sync = payload["sync"]
        in_queue = payload["queue"]
        ids_to_remove = []

        for out_entry in out_queue:
            if out_entry["recipient"] != sender:
                continue

            if out_entry["index"] <= sync:
                if out_entry["action"] == "resolve":
                    ids_to_remove.append(out_entry["id"])
                continue
            else:
                break

        new_in_index = in_index
        for in_entry in in_queue:
            if in_entry["recipient"] != data.id:
                continue

            index = in_entry["index"]
            if index <= in_index:
                continue

            if index == 0 and sync == 0:
                init_id = in_entry["id"]

            new_in_index = index
            await queue_in_entry(in_entry, ids_to_remove, data.portal)

        out_queue = [entry for entry in out_queue if entry["id"] not in ids_to_remove]

        state["queue"] = out_queue
        state["in_index"] = new_in_index
        state["init_id"] = init_id
        await persistence.set(data.id + "<>" + sender, dict(state))

        return None

    async def queue_in_entry(entry, ids_to_remove, portal):
        sender = entry["sender"]
        action = entry["action"]
        payload = entry["payload"]

        if action == "resolve":
            resolve_id = payload["id"]
            ids_to_remove.append(resolve_id)

            if resolve_id in futures:
                future = futures.pop(resolve_id, None)
                if future and not future.done():
                    if ("error" in payload) and payload["error"]:
                        future.set_exception(Exception(payload["error"]))
                    else:
                        future.set_result(payload["value"])
        else:
            res = None
            error = None
            try:
                res = await portal(
                    "local." + action,
                    {"sender": sender, "payload": payload},
                )
            except Exception as e:
                error = str(e)

            asyncio.create_task(
                portal(
                    "queue.out",
                    {
                        "recipient": sender,
                        "action": "resolve",
                        "payload": {"id": entry["id"], "value": res, "error": error},
                    },
                )
            )

    async def queue_dispatch(data, scope):
        persistence = scope.demand({"type": "persistence"})

        recipient = data.payload["recipient"]

        recipient_state_entry = await persistence.get(data.id + "<>" + recipient) or {}
        state = recipient_state_entry.get("value", {"queue": []})
        queue = state.get("queue", [])
        in_index = state.get("in_index", 0)

        data.portal(
            "send",
            {"recipient": recipient, "payload": {"sync": in_index, "queue": queue}},
        )

        return None

    return {
        "queue.in": queue_in,
        "queue.out": queue_out,
        "queue.dispatch": queue_dispatch,
    }



from .portals_lib import create_portal
from .local_portal import create_local_portal, create_local_ether
from .client_portal import create_client_portal, create_client_ether
from .server_portal import create_server_portal, create_server_ether
from .fetch_ether import create_fetch_ether
from .promised_value import promised_value
from .with_middleware import with_middleware
from .suppliers.crypto import random_uuid

__version__ = "0.0.7"



from .client_portal import create_client_ether

import asyncio


def create_fetch_ether(endpoint, fetch_method):
    async def send(data, signal):
        # print("Fetch Ether > send:", endpoint, data)

        if not isinstance(data, str):
            print(
                "fetch ether > send: unsupported data format. Only supporting 'string' at the moment."
            )
            return None

            # abort_controller = asyncio.Event()

            # def on_abort():
            #     abort_controller.set()

            # signal["on_abort"] = on_abort

        res = fetch_method(endpoint, data)

        if res.ok:
            return res.text
        else:
            raise Exception(
                f"Fetch Ether > send: request failed with status {res.status_code}"
            )

    return create_client_ether(send)



import unittest

import asyncio

import socket
from portals_lib import create_server_portal
from microenv import microenv


class TestServer(unittest.TestCase):
    def test_server(self):
        # def debug(data, scope):
        #     print(*data)

        server_portal = create_server_portal(
            microenv({"foo": "bar"}, {"id": "test_server"}),
            # {"debug": debug},
        )

        asyncio.run(server_portal("open"))

        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind(("localhost", 8000))
        server_socket.listen(1)

        print("Server is listening on port 8000...")

        try:
            while True:
                client_socket, addr = server_socket.accept()
                print(f"Connection from {addr}")

                try:
                    request = client_socket.recv(1024).decode("utf-8")
                    # print("Received request:", request)

                    if request.startswith("POST /portal"):
                        body_start = request.index("\r\n\r\n") + 4
                        body = request[body_start:]
                        print("Request body:", body)

                        async def handle_request():
                            response = await server_portal("receive", body)
                            client_socket.sendall(
                                f"HTTP/1.1 200 OK\r\nContent-Type: application/json\r\nContent-Length: {len(response)}\r\n\r\n{response}".encode(
                                    "utf-8"
                                )
                            )

                        asyncio.run(handle_request())
                    else:
                        client_socket.sendall(b"HTTP/1.1 404 Not Found\r\n\r\n")

                except Exception as e:
                    print(f"Error handling request: {e}")
                    client_socket.sendall(b"HTTP/1.1 500 Internal Server Error\r\n\r\n")
                finally:
                    client_socket.close()

        except KeyboardInterrupt:
            print("Server stopped by user.")
        finally:
            server_socket.close()


if __name__ == "__main__":
    unittest.main()



from .portals_lib import create_portal

import asyncio


def create_local_ether():
    _portals = {}

    async def ether_attach(data, scope):
        _portals[data.id] = data.portal

    async def ether_send(data, scope):
        payload = data.payload
        recipient = payload["recipient"]
        payload_to_receive = payload["payload"]

        while True:
            remote_portal = _portals.get(recipient)
            if not remote_portal:
                await asyncio.sleep(1)
                continue
            await remote_portal("receive", payload_to_receive)
            break

    async def ether_detach(data, scope):
        _portals.pop(data.id, None)

    return {
        "ether.attach": ether_attach,
        "ether.send": ether_send,
        "ether.detach": ether_detach,
    }


def create_local_portal(
    env,
    local_ether,
    middleware=None,
):
    return create_portal(env, [local_ether, middleware])



def not_implemented_supplier(data, scope):
    raise NotImplementedError(f"{scope.type} not implemented")


ether = {
    "ether.attach": not_implemented_supplier,
    "ether.send": not_implemented_supplier,
    "ether.detach": not_implemented_supplier,
}



from microenv import microenv


async def remote_enter_supplier(data, scope):
    remoteId = data.payload["id"]

    return await data.portal(
        "queue.out",
        {
            "recipient": remoteId,
            "action": "enter",
            "payload": {"id": remoteId},
        },
    )


async def remote_supplier(data, scope):
    action = scope.type.split(".")[1]
    recipient = data.payload["recipient"]
    payload = data.payload.copy()
    del payload["recipient"]

    return await data.portal(
        "queue.out",
        {
            "recipient": recipient,
            "action": action,
            "payload": payload,
        },
    )


remote = {
    "remote.enter": remote_enter_supplier,
    "remote.leave": remote_supplier,
    "remote.resolve": remote_supplier,
    "remote.get": remote_supplier,
    "remote.set": remote_supplier,
    "remote.call": remote_supplier,
}



async def receive(data, scope):
    received_data = data.portal("packager.unpack", data.payload)

    sender = received_data["sender"]

    await data.portal("queue.in", received_data)

    return {"sender": sender}



from .portals_lib import create_portal

import asyncio


def create_server_ether():
    resolvers_by_recipient = {}
    is_attached = [False]

    def middleware(super_methods):
        async def ether_attach(data, scope):
            is_attached[0] = True

        async def ether_detach(data, scope):
            is_attached[0] = False

        async def receive(data, scope):
            if not is_attached[0]:
                print("Server Ether -> tried to receive when detached")
                return

            portal = data.portal
            payload = data.payload

            receive_res = await portal("$super_receive", payload)

            sender = receive_res.get("sender") if receive_res else None
            if not sender:
                return

            loop = asyncio.get_event_loop()
            fut = loop.create_future()
            old_resolver = resolvers_by_recipient.get(sender)
            if old_resolver:
                old_resolver["resolve"](None)
            resolvers_by_recipient[sender] = {
                "resolve": fut.set_result,
                "reject": fut.set_exception,
            }

            return await fut

        async def ether_send(data, scope):
            if not is_attached[0]:
                print('Server Ether -> tried to "ether.send" when detached')
                return
            payload = data.payload
            recipient = payload["recipient"]
            payload_to_send = payload["payload"]
            resolver = resolvers_by_recipient.get(recipient)

            if resolver:
                resolvers_by_recipient.pop(recipient, None)
                resolver["resolve"](payload_to_send)

        return {
            "ether.attach": ether_attach,
            "ether.detach": ether_detach,
            "receive": receive,
            "ether.send": ether_send,
            "$super_receive": super_methods["receive"],
        }

    return middleware


def create_server_portal(
    env,
    middleware=None,
):
    return create_portal(env, [create_server_ether(), middleware])



def with_middleware(suppliers, middleware=None):
    result = suppliers.copy()
    if not middleware:
        return result
    mlist = middleware if isinstance(middleware, list) else [middleware]
    for m in mlist:
        if not m:
            continue
        overrides = m(result) if callable(m) else m
        for key, val in overrides.items():
            if val is not None:
                result[key] = val
    return result



try:
    import ujson as json
except ImportError:
    import json


def packager_pack(data, scope):
    return json.dumps(data.payload)


def packager_unpack(data, scope):
    return json.loads(data.payload)


json_packager = {
    "packager.pack": packager_pack,
    "packager.unpack": packager_unpack,
}



from supply_demand import supply_demand
from layered_persistence import LayeredPersistence, RuntimeLayer

from .with_middleware import with_middleware
from .suppliers import (
    init_supplier,
    ether_suppliers,
    local_suppliers,
    remote_suppliers,
    open_supplier,
    close_supplier,
    enter_supplier,
    leave_supplier,
    guest_supplier,
    crypto_supplier,
    json_packager,
    create_send_supplier,
    receive_supplier,
    create_queue_suppliers,
)


def create_portal(env, middleware=None):
    portal_id = env.descriptor.get("id", None)
    if not portal_id:
        raise ValueError("create_portal: given env has no id in descriptor")

    suppliers = {
        "init": init_supplier,
        "open": open_supplier,
        "close": close_supplier,
        "enter": enter_supplier,
        "leave": leave_supplier,
        "guest": guest_supplier,
        "send": create_send_supplier(),
        "receive": receive_supplier,
        "crypto": crypto_supplier,
    }

    suppliers.update(ether_suppliers)
    suppliers.update(local_suppliers)
    suppliers.update(remote_suppliers)
    suppliers.update(json_packager)
    suppliers.update(create_queue_suppliers({"outgoing_buffer_ms": 200}))

    # Apply middleware overrides
    suppliers = with_middleware(suppliers, middleware)

    # Ensure persistence supplier is present
    if "persistence" not in suppliers:
        persistence = LayeredPersistence([RuntimeLayer()])

        def persistenceSupplier(d, s):
            return persistence

        suppliers["persistence"] = persistenceSupplier

    def root_supplier(data, scope):
        # Invoke init supplier to establish portal
        return scope.demand(
            {
                "type": "init",
                "data": {"id": portal_id, "env": env},
            }
        )

    # Build portal method using supply_demand
    return supply_demand(root_supplier, suppliers)



import asyncio


def create_send_supplier():
    def send(data, scope):
        data_to_pack = dict(data.payload)
        data_to_pack["sender"] = data.id
        del data_to_pack["recipient"]

        send_payload = data.portal("packager.pack", data_to_pack)

        asyncio.create_task(
            data.portal(
                "ether.send",
                {"recipient": data.payload["recipient"], "payload": send_payload},
            )
        )

        return None

    return send



try:
    import inspect

    is_awaitable = inspect.isawaitable
except ImportError:
    # MicroPython
    def is_awaitable(val):
        is_awaitable = True
        try:
            return getattr(val, "__await__")
        except:
            try:
                getattr(val, "__next__")
            except:
                is_awaitable = False
        return is_awaitable


async def enter_supplier(data, scope):
    if data.id != data.payload["payload"]["id"]:
        raise ValueError("Unauthorised")

    data.portal("guest", data.payload["sender"])

    public_children = [
        child
        for child in data.env.descriptor["children"]
        if not child.get("private", False)
    ]

    env_descriptor = data.env.descriptor.copy()
    env_descriptor["children"] = public_children

    return {
        "id": data.id,
        "envDescriptor": env_descriptor,
    }


def local_supplier(data, scope):
    raise NotImplementedError(f"{scope.type} suppliers not implemented")


async def local_get_supplier(data, scope):
    env = data.env
    payload = data.payload["payload"]
    key = payload["key"]
    sender = data.payload.get("sender", None)
    next_ = data.payload.get("next", None)

    return env.get(key, sender, next_)


async def local_set_supplier(data, scope):
    env = data.env
    payload = data.payload["payload"]
    key = payload["key"]
    value = payload["value"]
    sender = data.payload.get("sender", None)

    return env.set(key, value, sender)


async def local_call_supplier(data, scope):
    env = data.env
    payload = data.payload["payload"]
    key = payload["key"]
    value = payload["value"]
    sender = data.payload.get("sender", None)

    res = env.get(key, sender)(value, env, sender)

    if is_awaitable(res):
        return await res

    return res


local = {
    "local.enter": enter_supplier,
    "local.leave": local_supplier,
    "local.get": local_get_supplier,
    "local.set": local_set_supplier,
    "local.call": local_call_supplier,
}



try:
    import urandom
    import ubinascii

    def uuid():
        random_bytes = bytearray(16)
        for i in range(16):
            random_bytes[i] = urandom.getrandbits(8)

        # Set version and variant bits:
        random_bytes[6] = (random_bytes[6] & 0x0F) | 0x40  # Version 4
        random_bytes[8] = (random_bytes[8] & 0x3F) | 0x80  # Variant 1

        hex_str = ubinascii.hexlify(random_bytes).decode()
        return "-".join(
            [hex_str[:8], hex_str[8:12], hex_str[12:16], hex_str[16:20], hex_str[20:]]
        )

except ImportError:
    from uuid import uuid4

    uuid = uuid4


def random_uuid():
    return str(uuid())


crypto = {"random_uuid": random_uuid}


def crypto_supplier(data, scope):
    return crypto



import unittest

import asyncio

from microenv import microenv

from portals_lib import (
    create_client_portal,
    create_server_portal,
)


class TestPortals(unittest.TestCase):
    def test_tmp(
        self,
    ):
        def debug(data, scope):
            return
            # print(*data)

        async def test():
            async def test_func(param, _ref=None, caller=None, next_=False):
                return {
                    "msg": "test_func called",
                    "param": param,
                }

            env_a = microenv(
                {"foo": "bar", "test_func": test_func},
                {"id": "envA"},
            )
            env_b = microenv({"foo": "not bar"}, {"id": "envB"})

            portal1 = create_server_portal(
                env_a,
                {"debug": debug},
            )
            portal2 = create_client_portal(
                env_b,
                lambda data, signal=None: portal1("receive", data),
                None,
                {"debug": debug},
            )

            # local_ether_middleware = create_local_ether()
            # portal1 = create_portal(env_a, {**local_ether_middleware, "debug": debug})
            # portal2 = create_portal(env_b, {**local_ether_middleware, "debug": debug})

            await portal1("open")
            env_a_via_portal = await portal2("enter", "envA")
            print("env_a_via_porta:", env_a_via_portal.descriptor)

            env_a_foo = await env_a_via_portal.face["foo"]
            print("env_a_foo:", env_a_foo)

            await env_a_via_portal.set("foo", "baz")
            env_a_foo2 = await env_a_via_portal.face["foo"]
            print("env_a_foo2:", env_a_foo2)

            print(
                "test_func call result:",
                await env_a_via_portal.face["test_func"]("test_param"),
            )

            self.assertIsNot(env_a_via_portal, None)

        asyncio.run(test())


if __name__ == "__main__":
    unittest.main()



from .portals_lib import create_portal

import asyncio


def get_timeout_from_attempts(num_attempts):
    return min(1000 * num_attempts, 10000)  # milliseconds


def create_client_ether(send, strategy=None):
    _portals = {}
    active_requests = [0]  # Mutable for nested update
    is_detached = [True]
    signal = [{}]

    async def ether_attach(data, scope):
        _portals[data.id] = data.portal
        is_detached[0] = False

    async def ether_send(data, scope):
        if signal[0].get("on_abort"):
            signal[0]["on_abort"]()
        signal[0] = {}
        if is_detached[0]:
            return False

        payload = data.payload
        portal = data.portal
        msg_data = payload["payload"]
        recipient = payload["recipient"]
        num_attempts = 0

        async def _send():
            if is_detached[0]:
                return None

            nonlocal num_attempts
            num_attempts += 1
            try:
                res = await send(msg_data, signal[0])
                signal[0] = {}
                if is_detached[0]:
                    return False
                return res
            except Exception as e:
                print(f"Client ether send error: {e}")
                abort_name = getattr(e, "name", "") or getattr(e, "args", [""])[0]
                if is_detached[0] or abort_name == "AbortError":
                    return False
                delay = get_timeout_from_attempts(num_attempts)
                await asyncio.sleep(delay)
                return await _send()

        active_requests[0] += 1
        received_data = await _send()
        active_requests[0] -= 1

        if received_data:
            await portal("receive", received_data)
        # if (
        #     strategy == "poll"
        #     and active_requests[0] == 0
        #     and received_data is not False
        # ):
        #     await portal("queue", ["dispatch", recipient])

    async def ether_detach(data, scope):
        is_detached[0] = True
        if signal[0].get("on_abort"):
            signal[0]["on_abort"]()
        signal[0] = {}

    return {
        "ether.attach": ether_attach,
        "ether.send": ether_send,
        "ether.detach": ether_detach,
    }


def create_client_portal(
    env,
    send,
    strategy=None,
    middleware=None,
):
    return create_portal(env, [create_client_ether(send, strategy), middleware])



from .init import init as init_supplier
from .ether import ether as ether_suppliers
from .local import local as local_suppliers
from .remote import remote as remote_suppliers
from .open import open as open_supplier
from .close import close as close_supplier
from .enter import enter as enter_supplier
from .leave import leave as leave_supplier
from .guest import guest as guest_supplier
from .crypto import crypto_supplier
from .json_packager import json_packager
from .send import create_send_supplier
from .receive import receive as receive_supplier
from .queue import create_queue_suppliers



class MiddlewareInput:
    def __init__(self, id, env, portal, payload):
        self.id = id
        self.env = env
        self.portal = portal
        self.payload = payload


def init(data, scope):
    id_ = data["id"]
    env = data["env"]
    demand = scope.demand

    has_debug = "debug" in scope.available_supplier_types
    depth = [1]

    def portal(action, payload=None):
        if has_debug:
            demand(
                {
                    "key": "debug",
                    "type": "debug",
                    "data": [f"[portal {id_}]", "··" * depth[0], action, payload],
                }
            )

            depth[0] += 1

            result = demand(
                {
                    "key": action,
                    "type": action,
                    "data": MiddlewareInput(
                        id_,
                        env,
                        portal,
                        payload,
                    ),
                }
            )

            depth[0] -= 1

            demand(
                {
                    "key": "debug",
                    "type": "debug",
                    "data": [
                        f"[portal {id_}]",
                        "··" * depth[0],
                        action,
                        "resolved as",
                        result,
                    ],
                }
            )
            return result
        else:
            return demand(
                {
                    "key": action,
                    "type": action,
                    "data": MiddlewareInput(
                        id_,
                        env,
                        portal,
                        payload,
                    ),
                }
            )

    return portal



from microenv import microenv


async def enter(data, scope):
    await data.portal("ether.attach")

    enter_data = await data.portal("remote.enter", {"id": data.payload})

    def remote_get(key, env, caller=None, next_=False):
        descriptor = env.descriptor
        property_descriptor = next(
            (child for child in descriptor["children"] if child["key"] == key), None
        )

        if (property_descriptor["type"] == "function") or (
            isinstance(property_descriptor["type"], dict)
            and property_descriptor["type"].get("extends") == "function"
        ):
            return lambda payload: data.portal(
                "remote.call",
                {"recipient": enter_data["id"], "key": key, "value": payload},
            )

        return data.portal(
            "remote.get",
            {"recipient": enter_data["id"], "key": key, "next": next_},
        )

    def remote_set(key, value, env, caller=None):
        descriptor = env.descriptor
        property_descriptor = next(
            (child for child in descriptor["children"] if child["key"] == key), None
        )

        if (property_descriptor["type"] == "function") or (
            isinstance(property_descriptor["type"], dict)
            and property_descriptor["type"].get("extends") == "function"
        ):
            raise ValueError("Cannot set function type")

        return data.portal(
            "remote.set",
            {
                "recipient": enter_data["id"],
                "key": key,
                "value": value,
            },
        )

    remote_env = microenv(
        {},
        enter_data["envDescriptor"],
        {
            "get": remote_get,
            "set": remote_set,
        },
    )

    return remote_env
